@@ -2,14 +2,18 @@ from __future__ import annotations
 
 import os
 import shlex
-from typing import Iterator, MutableMapping
+from collections.abc import MutableMapping
+from typing import TYPE_CHECKING
 
 import anyio
 
 from fastenv.utilities import logger
 
+if TYPE_CHECKING:
+    from collections.abc import Iterator
 
-class DotEnv(MutableMapping):
+
+class DotEnv(MutableMapping[str, str]):
     __slots__ = "_data", "source"
 
     def __init__(self, *args: str, **kwargs: str) -> None:
@@ -17,7 +21,7 @@ class DotEnv(MutableMapping):
         self.source: anyio.Path | list[anyio.Path] | None = None
         self.setenv(*args, **kwargs)
 
-    def __getitem__(self, key: str) -> str | None:
+    def __getitem__(self, key: str) -> str:
         return self._data[key]
 
     def __setitem__(self, key: str, value: str) -> None:
@@ -79,7 +83,7 @@ class DotEnv(MutableMapping):
         return tuple(
             (split_arg[0].strip(" \n\"'").upper(), split_arg[1].strip(" \n\"'"))
             for a in parsed_args
-            if len((split_arg := a.split(sep="=", maxsplit=1))) == 2
+            if len(split_arg := a.split(sep="=", maxsplit=1)) == 2
         )
 
     def _parse_kwargs(self, **kwargs: str) -> dict[str, str]:
